"""
coax.multiplexer
~~~~~~~~~~~~~~~~
"""

PORT_MAP_3299 = [
    # The 3299-2 port numbers appear to be LSB first
    0b000000,
    0b100000,
    0b010000,
    0b110000,
    0b001000,
    0b101000,
    0b011000,
    0b111000
]

def get_device_address(port):
    if port < 0 or port > 7:
        raise ValueError('Port must be between 0 and 7')

    return PORT_MAP_3299[port]



from .__about__ import __version__

from .interface import InterfaceFeature
from .serial_interface import SerialInterface, open_serial_interface

from .protocol import (
    PollAction,
    PollResponse,
    PowerOnResetCompletePollResponse,
    KeystrokePollResponse,
    TerminalType,
    Control,
    SecondaryControl,
    Poll,
    PollAck,
    ReadStatus,
    ReadTerminalId,
    ReadExtendedId,
    ReadAddressCounterHi,
    ReadAddressCounterLo,
    ReadData,
    ReadMultiple,
    Reset,
    LoadControlRegister,
    LoadSecondaryControl,
    LoadMask,
    LoadAddressCounterHi,
    LoadAddressCounterLo,
    WriteData,
    Clear,
    SearchForward,
    SearchBackward,
    InsertByte,
    StartOperation,
    DiagnosticReset,
    ReadFeatureId,
    EABReadData,
    EABLoadMask,
    EABWriteAlternate,
    EABReadMultiple,
    EABWriteUnderMask,
    EABReadStatus,
    Data
)

from .features import (
    Feature,
    read_feature_ids,
    parse_features
)

from .multiplexer import get_device_address

from .exceptions import (
    InterfaceError,
    ReceiveError,
    InterfaceTimeout,
    ReceiveTimeout,
    ProtocolError
)



"""
coax.protocol
~~~~~~~~~~~~~
"""

from enum import Enum

from .interface import FrameFormat
from .parity import odd_parity
from .exceptions import ProtocolError

class Command(Enum):
    """Terminal command."""

    # Base
    POLL = 0x01
    POLL_ACK = 0x11
    READ_STATUS = 0x0d
    READ_TERMINAL_ID = 0x09
    READ_EXTENDED_ID = 0x07
    READ_ADDRESS_COUNTER_HI = 0x05
    READ_ADDRESS_COUNTER_LO = 0x15
    READ_DATA = 0x03
    READ_MULTIPLE = 0x0b

    RESET = 0x02
    LOAD_CONTROL_REGISTER = 0x0a
    LOAD_SECONDARY_CONTROL = 0x1a
    LOAD_MASK = 0x16
    LOAD_ADDRESS_COUNTER_HI = 0x04
    LOAD_ADDRESS_COUNTER_LO = 0x14
    WRITE_DATA = 0x0c
    CLEAR = 0x06
    SEARCH_FORWARD = 0x10
    SEARCH_BACKWARD = 0x12
    INSERT_BYTE = 0x0e
    START_OPERATION = 0x08
    DIAGNOSTIC_RESET = 0x1c

    # Feature
    READ_FEATURE_ID = 0x07

    # EAB Feature
    EAB_READ_DATA = 0x03
    EAB_LOAD_MASK = 0x06
    EAB_WRITE_ALTERNATE = 0x0a
    EAB_READ_MULTIPLE = 0x0b
    EAB_WRITE_UNDER_MASK = 0x0c
    EAB_READ_STATUS = 0x0d

class PollAction(Enum):
    """Terminal POLL action."""

    NONE = 0x0
    ALARM = 0x2
    ENABLE_KEYBOARD_CLICKER = 0x3
    DISABLE_KEYBOARD_CLICKER = 0x1

class PollResponse:
    """Terminal POLL response."""

    @staticmethod
    def is_power_on_reset_complete(value):
        """Is the response word a power on reset complete response?"""
        return value == 0xa

    @staticmethod
    def is_keystroke(value):
        """Is the response word a keystroke response?"""
        return ((value & 0x2) == 0x2) and ((value & 0x1) == 0)

    def __init__(self, value):
        self.value = value

class PowerOnResetCompletePollResponse(PollResponse):
    """Terminal power-on-reset complete poll response."""

    def __init__(self, value):
        if not PollResponse.is_power_on_reset_complete(value):
            raise ValueError(f'Invalid POR poll response: {value}')

        super().__init__(value)

class KeystrokePollResponse(PollResponse):
    """Terminal keystroke poll response."""

    def __init__(self, value):
        if not PollResponse.is_keystroke(value):
            raise ValueError(f'Invalid keystroke poll response: {value}')

        super().__init__(value)

        self.scan_code = (value >> 2) & 0xff

    def __repr__(self):
        return f'<KeystrokePollResponse scan_code={self.scan_code}>'

class Status:
    """Terminal status."""

    def __init__(self, value):
        self.value = value

        self.monocase = bool(value & 0x80)
        self.busy = not bool(value & 0x20)
        self.feature_error = bool(value & 0x04)
        self.operation_complete = bool(value & 0x02)

    def __repr__(self):
        return (f'<Status monocase={self.monocase}, busy={self.busy}, '
                f'feature_error={self.feature_error}, '
                f'operation_complete={self.operation_complete}>')

class TerminalType(Enum):
    """Terminal type."""

    CUT = 1
    DFT = 2

class TerminalId:
    """Terminal model and keyboard."""

    _MODEL_MAP = {
        0b010: 2,
        0b011: 3,
        0b111: 4,
        0b110: 5
    }

    def __init__(self, value):
        self.value = value

        if (value & 0x1) == 0:
            self.type = TerminalType.CUT

            model = (value & 0x0e) >> 1

            if model not in TerminalId._MODEL_MAP:
                raise ValueError(f'Invalid model: {model}')

            self.model = TerminalId._MODEL_MAP[model]
            self.keyboard = (value & 0xf0) >> 4
        elif value == 1:
            self.type = TerminalType.DFT
            self.model = None
            self.keyboard = None
        else:
            raise ValueError(f'Invalid terminal identifier: {value}')

    def __repr__(self):
        return (f'<TerminalId type={self.type.name}, model={self.model}, '
                f'keyboard={self.keyboard}>')

class Control:
    """Terminal control register."""

    def __init__(self, step_inhibit=False, display_inhibit=False, cursor_inhibit=False,
                 cursor_reverse=False, cursor_blink=False):
        self.step_inhibit = step_inhibit
        self.display_inhibit = display_inhibit
        self.cursor_inhibit = cursor_inhibit
        self.cursor_reverse = cursor_reverse
        self.cursor_blink = cursor_blink

    @property
    def value(self):
        value = bool(self.step_inhibit) << 4
        value |= bool(self.display_inhibit) << 3
        value |= bool(self.cursor_inhibit) << 2
        value |= bool(self.cursor_reverse) << 1
        value |= bool(self.cursor_blink)

        return value

    def __repr__(self):
        return (f'<Control step_inhibit={self.step_inhibit}>, '
                f'display_inhibit={self.display_inhibit}, '
                f'cursor_inhibit={self.cursor_inhibit}, '
                f'cursor_reverse={self.cursor_reverse}, '
                f'cursor_blink={self.cursor_blink}>')

class SecondaryControl:
    """Terminal secondary control register."""

    def __init__(self, big=False):
        self.big = big

    @property
    def value(self):
        return int(bool(self.big))

    def __repr__(self):
        return f'<SecondaryControl big={self.big}>'

class ReadCommand:
    """Base class for read commands."""

    response_length = None

    def pack_outbound_frame(self):
        raise NotImplementedError

    def unpack_inbound_frame(self, words):
        raise NotImplementedError

class WriteCommand:
    """Base class for write commands."""

    response_length = 1

    def pack_outbound_frame(self):
        raise NotImplementedError

    def unpack_inbound_frame(self, words):
        if not is_tt_ar(words):
            raise ProtocolError(f'Expected TT/AR response: {words}')

class Poll(ReadCommand):
    """POLL command."""

    response_length = 1

    def __init__(self, action=PollAction.NONE):
        self.action = action

    def pack_outbound_frame(self):
        command_word = (self.action.value << 8) | pack_command_word(Command.POLL)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        if is_tt_ar(words):
            return None

        if len(words) != 1:
            raise ProtocolError(f'Expected 1 word POLL response: {words}')

        word = words[0]

        if PollResponse.is_power_on_reset_complete(word):
            return PowerOnResetCompletePollResponse(word)

        if PollResponse.is_keystroke(word):
            return KeystrokePollResponse(word)

        return PollResponse(word)

class PollAck(WriteCommand):
    """POLL_ACK command."""

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.POLL_ACK)

        return (FrameFormat.WORD_DATA, command_word)

class ReadStatus(ReadCommand):
    """READ_STATUS command."""

    response_length = 1

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.READ_STATUS)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        if len(words) != 1:
            raise ProtocolError(f'Expected 1 word READ_STATUS response: {words}')

        return Status(unpack_data_word(words[0]))

class ReadTerminalId(ReadCommand):
    """READ_TERMINAL_ID command."""

    response_length = 1

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.READ_TERMINAL_ID)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        if len(words) != 1:
            raise ProtocolError(f'Expected 1 word READ_TERMINAL_ID response: {words}')

        return TerminalId(unpack_data_word(words[0]))

class ReadExtendedId(ReadCommand):
    """READ_EXTENDED_ID command."""

    response_length = 4

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.READ_EXTENDED_ID)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        if is_tt_ar(words):
            return None

        if len(words) != 4:
            raise ProtocolError(f'Expected 4 word READ_EXTENDED_ID response: {words}')

        return unpack_data_words(words)

class ReadAddressCounterHi(ReadCommand):
    """READ_ADDRESS_COUNTER_HI command."""

    response_length = 1

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.READ_ADDRESS_COUNTER_HI)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        if len(words) != 1:
            raise ProtocolError(f'Expected 1 word READ_ADDRESS_COUNTER_HI response: {words}')

        return unpack_data_word(words[0])

class ReadAddressCounterLo(ReadCommand):
    """READ_ADDRESS_COUNTER_LO command."""

    response_length = 1

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.READ_ADDRESS_COUNTER_LO)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        if len(words) != 1:
            raise ProtocolError(f'Expected 1 word READ_ADDRESS_COUNTER_LO response: {words}')

        return unpack_data_word(words[0])

class ReadData(ReadCommand):
    """READ_DATA command."""

    response_length = 1

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.READ_DATA)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        if len(words) != 1:
            raise ProtocolError(f'Expected 1 word READ_DATA response: {words}')

        return unpack_data_word(words[0])

class ReadMultiple(ReadCommand):
    """READ_MULTIPLE command."""

    response_length = 32

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.READ_MULTIPLE)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        if len(words) == 0:
            raise ProtocolError(f'Expected 1 or more word READ_MULTIPLE response: {words}')

        return unpack_data_words(words)

class Reset(WriteCommand):
    """RESET command."""

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.RESET)

        return (FrameFormat.WORD_DATA, command_word)

class LoadControlRegister(WriteCommand):
    """LOAD_CONTROL_REGISTER command."""

    def __init__(self, control):
        self.control = control

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.LOAD_CONTROL_REGISTER)

        return (FrameFormat.WORD_DATA, command_word, [self.control.value])

class LoadSecondaryControl(WriteCommand):
    """LOAD_SECONDARY_CONTROL command."""

    def __init__(self, control):
        self.control = control

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.LOAD_SECONDARY_CONTROL)

        return (FrameFormat.WORD_DATA, command_word, [self.control.value])

class LoadMask(WriteCommand):
    """LOAD_MASK command."""

    def __init__(self, mask):
        self.mask = mask

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.LOAD_MASK)

        return (FrameFormat.WORD_DATA, command_word, [self.mask])

class LoadAddressCounterHi(WriteCommand):
    """LOAD_ADDRESS_COUNTER_HI command."""

    def __init__(self, address):
        if address < 0 or address > 255:
            raise ValueError('Address is out of range')

        self.address = address

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.LOAD_ADDRESS_COUNTER_HI)

        return (FrameFormat.WORD_DATA, command_word, [self.address])

class LoadAddressCounterLo(WriteCommand):
    """LOAD_ADDRESS_COUNTER_LO command."""

    def __init__(self, address):
        if address < 0 or address > 255:
            raise ValueError('Address is out of range')

        self.address = address

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.LOAD_ADDRESS_COUNTER_LO)

        return (FrameFormat.WORD_DATA, command_word, [self.address])

class WriteData(WriteCommand):
    """WRITE_DATA command."""

    def __init__(self, data):
        self.data = data

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.WRITE_DATA)

        return (FrameFormat.WORD_DATA, command_word, self.data)

class Clear(WriteCommand):
    """CLEAR command."""

    def __init__(self, pattern):
        self.pattern = pattern

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.CLEAR)

        return (FrameFormat.WORD_DATA, command_word, [self.pattern])

class SearchForward(WriteCommand):
    """SEARCH_FORWARD command."""

    def __init__(self, pattern):
        self.pattern = pattern

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.SEARCH_FORWARD)

        return (FrameFormat.WORD_DATA, command_word, [self.pattern])

class SearchBackward(WriteCommand):
    """SEARCH_BACKWARD command."""

    def __init__(self, pattern):
        self.pattern = pattern

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.SEARCH_BACKWARD)

        return (FrameFormat.WORD_DATA, command_word, [self.pattern])

class InsertByte(WriteCommand):
    """INSERT_BYTE command."""

    def __init__(self, byte):
        self.byte = byte

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.INSERT_BYTE)

        return (FrameFormat.WORD_DATA, command_word, [self.byte])

class StartOperation(WriteCommand):
    """START_OPERATION command."""

    def pack_outbound_frame(self):
        raise NotImplementedError

    def unpack_inbound_frame(self, words):
        raise NotImplementedError

class DiagnosticReset(WriteCommand):
    """DIAGNOSTIC_RESET command."""

    def pack_outbound_frame(self):
        raise NotImplementedError

    def unpack_inbound_frame(self, words):
        raise NotImplementedError

class ReadFeatureId(ReadCommand):
    """READ_FEATURE_ID command."""

    response_length = 1

    def __init__(self, feature_address):
        self.feature_address = feature_address

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.READ_FEATURE_ID, self.feature_address)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        if is_tt_ar(words):
            return None

        if len(words) != 1:
            raise ProtocolError(f'Expected 1 word READ_FEATURE_ID response: {words}')

        return unpack_data_word(words[0])

class EABReadData(ReadCommand):
    """EAB_READ_DATA command."""

    response_length = 1

    def __init__(self, feature_address):
        self.feature_address = feature_address

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.EAB_READ_DATA, self.feature_address)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        if len(words) != 1:
            raise ProtocolError(f'Expected 1 word EAB_READ_DATA response: {words}')

        return unpack_data_words(words)

class EABLoadMask(WriteCommand):
    """EAB_LOAD_MASK command."""

    def __init__(self, feature_address, mask):
        self.feature_address = feature_address
        self.mask = mask

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.EAB_LOAD_MASK, self.feature_address)

        return (FrameFormat.WORD_DATA, command_word, [self.mask])

class EABWriteAlternate(WriteCommand):
    """EAB_WRITE_ALTERNATE command."""

    def __init__(self, feature_address, data):
        self.feature_address = feature_address
        self.data = data

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.EAB_WRITE_ALTERNATE, self.feature_address)

        return (FrameFormat.WORD_DATA, command_word, self.data)

class EABReadMultiple(ReadCommand):
    """EAB_READ_MULTIPLE command."""

    response_length = 32

    def __init__(self, feature_address):
        self.feature_address = feature_address

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.EAB_READ_MULTIPLE, self.feature_address)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        return unpack_data_words(words)

class EABWriteUnderMask(WriteCommand):
    """EAB_WRITE_UNDER_MASK command."""

    def __init__(self, feature_address, byte):
        self.feature_address = feature_address
        self.byte = byte

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.EAB_WRITE_UNDER_MASK, self.feature_address)

        return (FrameFormat.WORD_DATA, command_word, [self.byte])

class EABReadStatus(ReadCommand):
    """EAB_READ_STATUS command."""

    response_length = 1

    def __init__(self, feature_address):
        self.feature_address = feature_address

    def pack_outbound_frame(self):
        command_word = pack_command_word(Command.EAB_READ_STATUS, self.feature_address)

        return (FrameFormat.WORD_DATA, command_word)

    def unpack_inbound_frame(self, words):
        if len(words) != 1:
            raise ProtocolError(f'Expected 1 word EAB_READ_STATUS response: {words}')

        return unpack_data_word(words[0])

class Data(WriteCommand):
    """Unaccompanied data."""

    def __init__(self, data):
        self.data = data

    def pack_outbound_frame(self):
        return (FrameFormat.DATA, self.data)

def pack_command_word(command, feature_address=None):
    """Pack a command into a 10-bit command word."""
    if feature_address is not None and (feature_address < 2 or feature_address > 15):
        raise ValueError(f'Invalid feature address: {feature_address}')

    return (feature_address << 6 if feature_address is not None else 0) | (command.value << 2) | 0x1

def pack_data_word(byte, set_parity=True):
    """Pack a data byte into a 10-bit data word."""
    parity = odd_parity(byte) if set_parity else 0

    return (byte << 2) | (parity << 1)

def is_tt_ar(words):
    """Is the word a TT/AR (transmission turnaround / auto response)?"""
    return len(words) == 1 and words[0] == 0

def is_data_word(word):
    """Is data word bit set?"""
    return (word & 0x1) == 0

def unpack_data_word(word, check_parity=False):
    """Unpack the data byte from a 10-bit data word."""
    if not is_data_word(word):
        raise ProtocolError(f'Word does not have data bit set: {word}')

    byte = (word >> 2) & 0xff
    parity = (word >> 1) & 0x1

    if check_parity and parity != odd_parity(byte):
        raise ProtocolError(f'Parity error: {word}')

    return byte

def pack_data_words(bytes_, set_parity=True):
    """Pack data bytes into 10-bit data words."""
    return [pack_data_word(byte, set_parity=set_parity) for byte in bytes_]

def unpack_data_words(words, check_parity=False):
    """Unpack the data bytes from 10-bit data words."""
    return bytes([unpack_data_word(word, check_parity=check_parity) for word in words])



"""
coax.serial_interface
~~~~~~~~~~~~~~~~~~~~~
"""

import time
import os
import struct
from copy import copy
from contextlib import contextmanager
from serial import Serial
from sliplib import SlipWrapper, ProtocolError

from .interface import Interface, InterfaceFeature, FrameFormat
from .protocol import pack_data_word
from .exceptions import InterfaceError, InterfaceTimeout, ReceiveError, ReceiveTimeout

class SerialInterface(Interface):
    """Serial attached 3270 coax interface."""

    def __init__(self, serial):
        if serial is None:
            raise ValueError('Serial port is required')

        super().__init__()

        self.serial = serial

        self.slip_serial = SlipSerial(self.serial)

        self.legacy_firmware_detected = None
        self.legacy_firmware_version = None

    def reset(self):
        """Reset the interface."""
        original_serial_timeout = self.serial.timeout

        self.serial.timeout = 5

        self.serial.reset_input_buffer()

        self._write_message(bytes([0x01]))

        try:
            message = self._read_message()
        finally:
            self.serial.timeout = original_serial_timeout

            self.serial.reset_input_buffer()

        if message[0] != 0x01:
            raise _convert_error(message)

        if message[1:] == b'\x32\x70':
            self.legacy_firmware_detected = False
            self.legacy_firmware_version = None
        elif len(message) == 4:
            (major, minor, patch) = struct.unpack('BBB', message[1:])

            self.legacy_firmware_detected = True
            self.legacy_firmware_version = f'{major}.{minor}.{patch}'
        else:
            raise InterfaceError(f'Invalid reset response: {message}')

        # Query features, if this is not a legacy firmware.
        if not self.legacy_firmware_detected:
            try:
                self.features = self._get_features()
            except InterfaceError:
                pass

    def enter_dfu_mode(self):
        """Enter device firmware upgrade mode."""
        message = bytes([0xf2])

        self._write_message(message)

        message = self._read_message()

        if message[0] != 0x01:
            raise _convert_error(message)

    def _get_features(self):
        """Get interface features."""
        message = bytes([0xf0, 0x07])

        self._write_message(message)

        message = self._read_message()

        if message[0] != 0x01:
            return _convert_error(message)

        known_feature_values = {feature.value for feature in InterfaceFeature}

        features = {InterfaceFeature(value) for value in message[1:] if value in known_feature_values}

        return features

    def _transmit_receive(self, outbound_frames, response_lengths, timeout):
        if len(response_lengths) != len(outbound_frames):
            raise ValueError('Response lengths length must equal outbound frames length')

        if any(address is not None for (address, _) in outbound_frames) and InterfaceFeature.PROTOCOL_3299 not in self.features:
            raise NotImplementedError('Interface does not support 3299 protocol')

        # Pack all messages before sending.
        timeout_milliseconds = self._calculate_timeout_milliseconds(timeout)

        messages = [_pack_transmit_receive_message(address, frame, response_length, timeout_milliseconds)
                    for ((address, frame), response_length) in zip(outbound_frames, response_lengths)]

        responses = []

        for message in messages:
            self._write_message(message)

            message = self._read_message()

            if message[0] == 0x01:
                response = _unpack_transmit_receive_response(message[1:])
            else:
                error = _convert_error(message)

                if not isinstance(error, (ReceiveError, ReceiveTimeout)):
                    raise error

                response = error

            responses.append(response)

        return responses

    def _calculate_timeout_milliseconds(self, timeout):
        milliseconds = 0

        if timeout:
            if self.serial.timeout and timeout > self.serial.timeout:
                raise ValueError('Timeout cannot be greater than serial timeout')

            milliseconds = int(timeout * 1000)

        return milliseconds

    def _read_message(self):
        try:
            message = self.slip_serial.recv_msg()
        except ProtocolError:
            raise InterfaceError('SLIP protocol error')

        if len(message) < 4:
            raise InterfaceError(f'Invalid response message: {message}')

        (length,) = struct.unpack('>H', message[:2])

        if length != len(message) - 4:
            raise InterfaceError('Response message length mismatch')

        if length < 1:
            raise InterfaceError('Empty response message')

        return message[2:-2]

    def _write_message(self, message):
        self.slip_serial.send_msg(struct.pack('>H', len(message)) + message +
                                  struct.pack('>H', 0))

@contextmanager
def open_serial_interface(serial_port, reset=True):
    """Opens serial port and initializes serial attached 3270 coax interface."""
    with Serial(serial_port, 115200) as serial:
        serial.reset_input_buffer()
        serial.reset_output_buffer()

        # Allow the interface firmware time to start, this is only required for the
        # original Arduino Mega based interface.
        if 'COAX_FAST_START' not in os.environ:
            time.sleep(3)

        interface = SerialInterface(serial)

        if reset:
            interface.reset()

        yield interface

def _pack_transmit_receive_message(address, frame, response_length, timeout_milliseconds):
    message = bytes([0x06])

    repeat_count = 0
    repeat_offset = 0
    bytes_ = bytearray()

    if frame[0] == FrameFormat.WORDS:
        if isinstance(frame[1], tuple):
            repeat_count = frame[1][1]

            for word in frame[1][0]:
                bytes_ += struct.pack('<H', word)
        else:
            for word in frame[1]:
                bytes_ += struct.pack('<H', word)
    elif frame[0] == FrameFormat.WORD_DATA:
        bytes_ += struct.pack('<H', frame[1])

        if len(frame) > 2:
            if isinstance(frame[2], tuple):
                repeat_offset = 1
                repeat_count = frame[2][1]

                for byte in frame[2][0]:
                    bytes_ += struct.pack('<H', pack_data_word(byte))
            else:
                for byte in frame[2]:
                    bytes_ += struct.pack('<H', pack_data_word(byte))
    elif frame[0] == FrameFormat.DATA:
        if isinstance(frame[1], tuple):
            repeat_count = frame[1][1]

            for byte in frame[1][0]:
                bytes_ += struct.pack('<H', pack_data_word(byte))
        else:
            for byte in frame[1]:
                bytes_ += struct.pack('<H', pack_data_word(byte))

    if address is not None:
        if address < 0 or address > 63:
            raise ValueError('Address must be between 0 and 63')

        if repeat_count > 0:
            repeat_offset += 1

        bytes_ = struct.pack('<H', 0x8000 | address) + bytes_

    message += struct.pack('>H', (repeat_offset << 15) | repeat_count)
    message += bytes_
    message += struct.pack('>HH', response_length, timeout_milliseconds)

    return message

def _unpack_transmit_receive_response(bytes_):
    return [(hi << 8) | lo for (lo, hi) in zip(bytes_[::2], bytes_[1::2])]

ERROR_MAP = {
    1: InterfaceError('Invalid request message'),
    2: InterfaceError('Unknown command'),

    101: InterfaceError('Receiver active'),
    102: ReceiveTimeout(),
    103: ReceiveError('Receiver buffer overflow'),
    104: ReceiveError('Receiver error')
}

def _convert_error(message):
    if message[0] != 0x02:
        return InterfaceError(f'Invalid response: {message}')

    if len(message) < 2:
        return InterfaceError(f'Invalid error response: {message}')

    if message[1] in ERROR_MAP:
        error = copy(ERROR_MAP[message[1]])

        # Append description if included.
        if len(message) > 2:
            description = message[2:].decode('ascii')

            if error.args:
                error.args = (f'{error.args[0]}: {description}', *error.args[1:])
            else:
                error.args = (description,)

        return error

    return InterfaceError(f'Unknown error: {message[1]}')

class SlipSerial(SlipWrapper):
    """sliplib wrapper for pySerial."""

    def send_bytes(self, packet):
        """Sends a packet over the serial port."""
        self.stream.write(packet)
        self.stream.flush()

    def recv_bytes(self):
        """Receive data from the serial port."""
        if self.stream.closed:
            return b''

        count = self.stream.in_waiting

        if count:
            return self.stream.read(count)

        byte = self.stream.read(1)

        if byte == b'':
            raise InterfaceTimeout()

        return byte



"""
coax.parity
~~~~~~~~~~~

Single byte parity computation
"""

# From http://p-nand-q.com/python/algorithms/math/bit-parity.html
def _parallel_swar(i):
    i = i - ((i >> 1) & 0x55555555)
    i = (i & 0x33333333) + ((i >> 2) & 0x33333333)
    i = (((i + (i >> 4)) & 0x0F0F0F0F) * 0x01010101) >> 24
    return int(i % 2)

_PARITY_LOOKUP = [_parallel_swar(i) for i in range(256)]

def even_parity(byte):
    """Compute even parity"""
    if byte < 0 or byte > 255:
        raise ValueError('Input must be between 0 and 255')

    return _PARITY_LOOKUP[byte]

def odd_parity(byte):
    """Compute odd parity"""
    return int(not even_parity(byte))



"""
coax.interface
~~~~~~~~~~~~~~
"""

from enum import Enum

from .exceptions import ProtocolError

class Interface:
    """3270 coax interface."""

    def __init__(self):
        self.features = set()

    def reset(self):
        """Reset the interface."""
        raise NotImplementedError

    def execute(self, commands, timeout=None):
        """Execute one or more commands."""
        (normalized_commands, has_multiple_commands) = _normalize_commands(commands)

        responses = self._execute(normalized_commands, timeout)

        if has_multiple_commands:
            return responses

        response = responses[0]

        if isinstance(response, BaseException):
            raise response

        return response

    def _execute(self, commands, timeout):
        (outbound_frames, response_lengths) = _pack_outbound_frames(commands)

        inbound_frames = self._transmit_receive(outbound_frames, response_lengths, timeout)

        responses = _unpack_inbound_frames(inbound_frames, commands)

        return responses

    def _transmit_receive(self, outbound_frames, response_lengths, timeout):
        raise NotImplementedError

class InterfaceFeature(Enum):
    """Interface feature."""

    PROTOCOL_3299 = 0x10

class FrameFormat(Enum):
    """3270 coax frame format."""

    WORDS = 1 # 10-bit words
    WORD_DATA = 2 # 10-bit word, 8-bit data words
    DATA = 4 # 8-bit data words

def _is_command(command):
    return hasattr(command, 'pack_outbound_frame') and hasattr(command, 'unpack_inbound_frame')

def _normalize_command(command):
    if _is_command(command):
        return (None, command)

    if not isinstance(command, tuple):
        raise TypeError('Invalid command form')

    if len(command) != 2:
        raise TypeError('Invalid command form')

    if not (command[0] is None or isinstance(command[0], int)):
        raise TypeError('Invalid command form')

    if not _is_command(command[1]):
        raise TypeError('Invalid command form')

    return command

def _normalize_commands(commands):
    try:
        command = _normalize_command(commands)

        return ([command], False)
    except TypeError:
        pass

    try:
        commands = list(commands)
    except TypeError:
        raise TypeError('Commands must be valid command or iterable')

    if not commands:
        raise ValueError('Commands must not be empty')

    return ([_normalize_command(command) for command in commands], True)

def _pack_outbound_frames(commands):
    frames = []
    response_lengths = []

    for (address, command) in commands:
        frame = command.pack_outbound_frame()
        response_length = command.response_length or 1

        frames.append((address, frame))
        response_lengths.append(response_length)

    return (frames, response_lengths)

def _unpack_inbound_frames(frames, commands):
    responses = []

    for (frame, (_, command)) in zip(frames, commands):
        if isinstance(frame, BaseException):
            responses.append(frame)
        else:
            try:
                response = command.unpack_inbound_frame(frame)
            except ProtocolError as error:
                response = error

            responses.append(response)

    return responses



"""
coax.features
~~~~~~~~~~~~~
"""

from enum import Enum

from .protocol import ReadFeatureId

class Feature(Enum):
    """Terminal feature."""

    EAB = 0x79

FEATURE_ADDRESS_MIN = 2
FEATURE_ADDRESS_MAX = 15

FEATURE_ADDRESSES = range(FEATURE_ADDRESS_MIN, FEATURE_ADDRESS_MAX + 1)

def read_feature_ids(addresses=None):
    """Generate READ_FEATURE_ID commands."""
    return [ReadFeatureId(address) for address in addresses or FEATURE_ADDRESSES]

def parse_features(ids, commands):
    """Parse READ_FEATURE_ID command responses into a map of features and addresses."""
    addresses = [command.feature_address for command in commands]

    known_ids = {feature.value for feature in Feature}

    features = {}

    for (address, id_) in zip(addresses, ids):
        if id_ is not None and id_ in known_ids:
            feature = Feature(id_)

            features[feature] = address

    return features
